import numpy as np
from sklearn import datasets
from sklearn.metrics import confusion_matrix
import seaborn as sns
import matplotlib.pyplot as plt

def relu(t):
    return np.maximum(t, 0)

def softmax(t):
    out = np.exp(t)
    return out / np.sum(out)

def predict(x):
    t1 = x @ W1 + b1
    h1 = relu(t1)
    t2 = h1 @ W2 + b2
    return t2

INPUT_DIM = 4
OUT_DIM = 3
H_DIM = 3

with open("w1.npy", 'rb') as temp:
    W1=np.load(temp)
with open("b1.npy", 'rb') as temp:
    b1=np.load(temp)
with open("w2.npy", 'rb') as temp:
    W2=np.load(temp)
with open("b2.npy", 'rb') as temp:
    b2=np.load(temp)

iris = datasets.load_iris()
dataset = []
for i in range(len(iris.target)):
    dataset.append((iris.data[[i]], iris.target[i]))

batch_x, batch_y = zip(*dataset)
xd = np.concatenate(batch_x, axis=0)
yd = np.array(batch_y)

predictions = predict(xd)
predicted_classes = np.argmax(predictions, axis=1)

conf_mat = confusion_matrix(yd, predicted_classes)
class_names = ['Setosa', 'Versicolor', 'Virginica']

plt.figure(figsize=(8, 6))
sns.heatmap(conf_mat, annot=True, fmt="d", cmap="Blues", xticklabels=class_names, yticklabels=class_names)
plt.ylabel('True label')
plt.xlabel('Predicted label')
plt.show()