import csv
import random
import numpy as np

np.random.seed(42)

INPUT_DIM = 4
OUT_DIM = 3
H_DIM = 3
ALPHA = 0.0001
NUM_EPOCHS = 10000
BATCH_SIZE = 50

def relu(t):
    return np.maximum(t, 0)

def softmax(t):
    out = np.exp(t)
    return out / np.sum(out)

def softmax_batch(t):
    return np.exp(t) / np.sum(np.exp(t), axis=1, keepdims=True)

def sparse_cross_entropy(z, y):
    return -np.log(z[0, y])

def sparse_cross_entropy_batch(z, y):
    return -np.log(np.array([z[j, y[j]] for j in range(len(y))]))

def to_full_batch(y, num_classes):
    return np.eye(num_classes)[y]

def relu_deriv(t):
    return np.where(t > 0, 1, 0)

from sklearn import datasets
iris = datasets.load_iris()

dataset = []
for i in range(len(iris.target)):
    dataset.append((iris.data[[i]], iris.target[i]))

loss_arr = []

bad = True
while bad:
    W1 = np.random.randn(INPUT_DIM, H_DIM) * np.sqrt(2/INPUT_DIM)
    b1 = np.zeros((1, H_DIM))
    W2 = np.random.randn(H_DIM, OUT_DIM) * np.sqrt(2/H_DIM)
    b2 = np.zeros((1, OUT_DIM))

    for ep in range(NUM_EPOCHS):
        random.shuffle(dataset)
        for i in range(len(dataset) // BATCH_SIZE):

            batch_x, batch_y = zip(*dataset[i*BATCH_SIZE : i*BATCH_SIZE+BATCH_SIZE])
            x = np.concatenate(batch_x, axis=0)
            y = np.array(batch_y)

            # Forward
            t1 = x @ W1 + b1
            h1 = relu(t1)
            t2 = h1 @ W2 + b2
            z = softmax_batch(t2)
            E = np.sum(sparse_cross_entropy_batch(z, y))

            # Backward
            y_full = to_full_batch(y, OUT_DIM)
            dE_dt2 = z - y_full
            dE_dW2 = h1.T @ dE_dt2
            dE_db2 = np.sum(dE_dt2, axis=0, keepdims=True)
            dE_dh1 = dE_dt2 @ W2.T
            dE_dt1 = dE_dh1 * relu_deriv(t1)
            dE_dW1 = x.T @ dE_dt1
            dE_db1 = np.sum(dE_dt1, axis=0, keepdims=True)

            # Update
            W1 -= ALPHA * dE_dW1
            b1 -= ALPHA * dE_db1
            W2 -= ALPHA * dE_dW2
            b2 -= ALPHA * dE_db2

            loss_arr.append(E)

    def predict(x):
        t1 = x @ W1 + b1
        h1 = relu(t1)
        t2 = h1 @ W2 + b2
        z = softmax_batch(t2)
        return z

    def calc_accuracy():
        correct = 0
        for x, y in dataset:
            z = predict(x)
            y_pred = np.argmax(z)
            if y_pred == y:
                correct += 1
        acc = correct / len(dataset)
        return acc

    accuracy = calc_accuracy()
    print("Accuracy:", accuracy)
    if accuracy >= 0.99:
        bad = False
        np.save('w1', W1)
        np.save('b1', b1)
        np.save('w2', W2)
        np.save('b2', b2)


import matplotlib.pyplot as plt
plt.plot(loss_arr)
plt.show()